#!/bin/bash
#
#  build-initrd.sh
#
#  Written by Jari Ruusu, July 16 2002
#
#  Copyright 2001,2002 by Jari Ruusu.
#  Redistribution of this file is permitted under the GNU Public License.
#
#  Changes by Hauke Johannknecht <ash@ash.de> 11/2001
#     - added Pivot mode
#     - added conffile-loading
#     - added initrdonly-mode
#     - added device 0 nodes
#
#
#  kernel .config :  CONFIG_BLK_DEV_RAM=y
#                    CONFIG_BLK_DEV_RAM_SIZE=4096
#                    CONFIG_BLK_DEV_INITRD=y
#                    CONFIG_MINIX_FS=y
#                    CONFIG_PROC_FS=y
#
#  /etc/lilo.conf :  initrd=/boot/initrd.gz
#  (USEPIVOT=0)      root=/dev/ram1
#                                 ^
#  /etc/lilo.conf :  append="init=/linuxrc"
#  (USEPIVOT=1)      initrd=/boot/initrd.gz
#                    root=/dev/ram0
#                                 ^
#  /boot/grub/menu.lst :  root (hd0,0)
#  (USEPIVOT=0)           kernel /vmlinuz ro root=/dev/ram1
#                         initrd /initrd.gz               ^
#
#  /boot/grub/menu.lst :  root (hd0,0)
#  (USEPIVOT=1)           kernel /vmlinuz root=/dev/ram0 init=/linuxrc
#                         initrd /initrd.gz            ^
#
#  root=XXX and append="root=XXX" are NOT equivalent. devfs+lilo users should
#  substitute root=/dev/ram0 with root=/dev/rd/0 . devfs+grub users should not
#  specify root= at all, but use rdev utility to directly write root device to
#  kernel image. Something like this:  rdev /boot/vmlinuz /dev/rd/0
#
#  usage :  ./build-initrd.sh [configfile]
#           lilo
#           cp -p /lib/modules/`uname -r`/block/loop.o /boot/loop-`uname -r`.o
#

### All default-values can be altered via the configfile

# normal /boot partition
BOOTDEV=/dev/hda1

# /boot partition type
BOOTTYPE=ext2

# encrypted root partition
CRYPTROOT=/dev/hda2

# root partition type
ROOTTYPE=ext2

# encryption type (AES128 / AES192 / AES256) of root partition
CIPHERTYPE=AES128

# optional password seed for root partition
#PSEED="-S XXXXXX"

# 1 = use GnuPG key to mount root partition, 0 = use normal key.
# If this is enabled (USEGPGKEY=1), file named rootkey.gpg (or
# whatever GPGKEYFILE is set to) must be manually copied to /boot.
# If rootkey.gpg is not encrypted with symmetric cipher, pubring.gpg
# and secring.gpg must be manually copied to /boot. Try to keep
# pubring.gpg as small as possible, because trustdb.gpg must fit
# into ramdisk which has very limited space available.
USEGPGKEY=0

# GnuPG key filename. Only used if USEGPGKEY=1
GPGKEYFILE=rootkey.gpg

# 1 = use loop module, 0 = loop driver linked to kernel
USEMODULE=1

# 1 = use pivot_root, 0 = use old change_root
# See above header for root= and append="init=" lilo.conf definitions.
# pivot_root is not available on 2.2 and 2.0 kernels.
USEPIVOT=0

# 1 = stop after creating and copying initrd, 0 = also copy tools/libs
INITRDONLY=0

# source root directory where files are copied from
SOURCEROOT=

# destination root directory where files are written to
DESTINATIONROOT=

# dest-dir below dest-root
DESTINATIONPREFIX=/boot

# name of created init ram-disk
INITRDGZNAME=initrd.gz

# Encrypted root loop device index (0 ... 7), 5 == /dev/loop5
# Device index must be one character even if max_loop is greater than 8
# _must_ match /etc/fstab entry:   /dev/loop5  /  ext2  defaults,xxxx  0  1
ROOTLOOPINDEX=5

# Temporary loop device index used in this script, 7 == /dev/loop7
TEMPLOOPINDEX=7

# Additional loop module parameters.
# Example: LOOPMODPARAMS="max_loop=8 lo_prealloc=125,5,200"
LOOPMODPARAMS=""

# 1 = use devfs, 0 = use classic disk-based device names. If this is
# enabled (USEDEVFS=1) then setting USEPIVOT=1 is also required.
USEDEVFS=0

# 1 = load national keyboard layout, 0 = don't load
# You _must_ manually copy correct keyboard layout to /boot/default.kmap
# which must be in uncompressed form. (can not be .gz file)
LOADNATIONALKEYB=0

# 1 = use "rootsetup" program that executes losetup to initialize loop
# 0 = use normal "losetup" program directly to initialize loop
# If enabled, rootsetup program (+libs) _must_ be manually copied to /boot.
USEROOTSETUP=0


if [ $# = 1 ] ; then
    if [ ! -f $1 ] ; then
        echo "ERROR: Can't find configfile '$1'"
        echo "Usage: $0 [configfile]"
        exit 1
    fi
    echo "Loading config from '$1'"
    . $1
fi

DEVFSSLASH1=
DEVFSRAMDSK=/dev/ram
if [ ${USEDEVFS} == 1 ] ; then
    DEVFSSLASH1=/
    DEVFSRAMDSK=/dev/rd/
fi
DEVFSSLASH2=
if [ -c /dev/.devfsd ] ; then
    DEVFSSLASH2=/
fi
LOSETUPPROG=losetup
if [ ${USEROOTSETUP} == 1 ] ; then
    LOSETUPPROG=rootsetup
fi

set -e
umask 077
cat - <<EOF >tmp-c-$$.c

/* Note: this program does not initialize C library, so all level 3     */
/* library calls are forbidden. Only level 2 systems calls are allowed. */

#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/mount.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include <fcntl.h>

#if ${USEPIVOT}
# ifdef __ia64__
#  include <sys/syscall.h>
#  define pivot_root(new_root,put_old) syscall(SYS_pivot_root,new_root,put_old)
# else
#  include <linux/unistd.h>
#  if !defined(__NR_pivot_root) && defined(__i386__)
#   define __NR_pivot_root 217
#  endif
static _syscall2(int,pivot_root,const char *,new_root,const char *,put_old)
# endif
#endif

#if defined __GLIBC__ && __GLIBC__ >= 2
# include <sys/reboot.h>
# define HaltKernel() reboot(0xCDEF0123)  /* LINUX_REBOOT_CMD_HALT */
#else
extern int reboot(int, int, int);
# define HaltKernel() reboot(0xfee1dead, 672274793, 0xCDEF0123)
#endif

static char * argv_init[] = { "init", 0, };
static char * envp_init[] = { "HOME=/", "TERM=linux", 0, };

void strCat(char *d, char *s)
{
    while(*d) d++;
    while(*s) *d++ = *s++;
    *d = 0;
}

void wrStr(char *s)
{
    char *p = s;
    int x = 0;

    while(*p) p++, x++;
    write(1, s, x);
}

int exeWait(char *p)
{
    int x, y;
    char *a[50], *e[1];

    if(!(x = fork())) {
        while(*p && (x < ((sizeof(a) / sizeof(char *)) - 1))) {  
            a[x++] = p;
            while(*p && (*p != ' ') && (*p != '\t')) p++;
            while((*p == ' ') || (*p == '\t')) *p++ = 0;
        }
        e[0] = a[x] = 0;
        if(x) execve(a[0], &a[0], &e[0]);
        _exit(1); 
    }
    if(x == -1) {
        wrStr("linuxrc: fork failed\n");
        return(1);
    }
    waitpid(x, &y, 0);
    if(!WIFEXITED(y) || (WEXITSTATUS(y) != 0)) {
        wrStr("Command \""); wrStr(p); wrStr("\" returned error\n");
        return(1);
    }
    return(0);
}

void doHalt()
{
    int x, y;

    sync();
    if(!(x = fork())) HaltKernel();
    waitpid(x, &y, 0);
    for(;;) ;
}

void runInit()
{
    execve("/sbin/init", argv_init, envp_init);
    execve("/etc/init", argv_init, envp_init);
    execve("/bin/init", argv_init, envp_init);
    wrStr("Bummer! Launching init failed\n");
    doHalt();
}

void _start()
{
    int x = 0;
#if ${USEMODULE}
    struct utsname un;
#endif
    char buf[300];

#if ${USEPIVOT}
    if(getpid() != 1) {
        /* pivot_root was configured, but kernel has */
        /* wandered off to change_root code path!    */
        _exit(0);
    }
#else    
    if(getpid() == 1) {
        /* change_root was configured, but kernel    */
        /* has wandered off to pivot_root code path! */
        runInit();
    }
#endif

    /* this intentionally mounts /boot partition as /lib */
    if(mount("${BOOTDEV}", "/lib", "${BOOTTYPE}", MS_MGC_VAL | MS_RDONLY, 0)) {
        wrStr("Mounting ${BOOTDEV} as /lib failed\n");
        goto fail4;
    }

#if ${LOADNATIONALKEYB}
    buf[0] = 0;
    strCat(buf, "/lib/loadkeys /lib/default.kmap");
    exeWait(buf);
#endif

#if ${USEMODULE}
    uname(&un);
    buf[0] = 0;
    strCat(buf, "/lib/insmod -o loop /lib/loop-");
    strCat(buf, &un.release[0]);
    strCat(buf, ".o ${LOOPMODPARAMS}");
    if(exeWait(buf)) goto fail3;
#endif

#if ${USEGPGKEY} && ${USEPIVOT}
    /* gpg needs a writable ramdisk, so remount rw */
    mount("${DEVFSRAMDSK}0", "/", "minix", MS_MGC_VAL | MS_REMOUNT, 0);
#endif

    tryAgain:
#if !(${USEROOTSETUP})
    wrStr("\nEncrypted file system, please supply correct password to continue\n\n");
#endif

    buf[0] = 0;
    strCat(buf, "/lib/${LOSETUPPROG} -e ${CIPHERTYPE} ${PSEED}");
#if ${USEGPGKEY}
    strCat(buf, " -K /lib/${GPGKEYFILE} -G /");
    unlink("/trustdb.gpg");
#endif
    strCat(buf, " /dev/loop${DEVFSSLASH1}${ROOTLOOPINDEX} ${CRYPTROOT}");

    if(exeWait(buf)) {
        if(++x >= 5) goto fail3;
        goto tryAgain;
    }
#if ${USEGPGKEY}
    unlink("/trustdb.gpg");
#endif

#if !(${USEROOTSETUP})
    wrStr("\n");
#endif

#if ${USEPIVOT}
    if(mount("/dev/loop${DEVFSSLASH1}${ROOTLOOPINDEX}", "/new-root", "${ROOTTYPE}", MS_MGC_VAL | MS_RDONLY, 0)) {
        wrStr("Looks like you didn't say the magic word. Mounting /dev/loop${DEVFSSLASH1}${ROOTLOOPINDEX} failed\n");
        buf[0] = 0;
        strCat(buf, "/lib/${LOSETUPPROG} -d /dev/loop${DEVFSSLASH1}${ROOTLOOPINDEX}");
        if(exeWait(buf)) goto fail3;
        if(++x >= 5) goto fail3;
        goto tryAgain;
    }
    umount("/lib");
    if(chdir("/new-root")) {
        wrStr("chdir() to /new-root failed\n");
        goto fail1;
    }
    if(pivot_root(".", "initrd")) {
        wrStr("pivot_root() to new root failed. Older kernels don't have pivot_root()\n");
        fail1:
        chdir("/");
        umount("/new-root");
        goto fail4;
    }
    chdir("/");

#if ${USEDEVFS}
    if(mount("none", "dev", "devfs", MS_MGC_VAL, 0)) {
        wrStr("Mounting /dev failed\n");
        goto fail1;
    }
#endif

    close(0);
    close(1);
    close(2);
    open("dev/console", O_RDWR, 0);
    dup(0);
    dup(0);

    if(chroot(".")) {
        wrStr("chroot() to new root failed\n");
        goto fail1;
    }
    wrStr("Pivoting to encrypted root completed successfully\n");
    runInit();
#else
    if(mount("none", "/proc", "proc", MS_MGC_VAL, 0)) {
        wrStr("Mounting /proc failed\n");
        goto fail3;
    }
    if((x = open("/proc/sys/kernel/real-root-dev", O_WRONLY, 0)) == -1) {
        wrStr("Unable to open real-root-dev\n");
        goto fail2;
    }
    write(x, "0x70${ROOTLOOPINDEX}\n", 6);
    close(x);
    fail2:
    umount("/proc");
#endif

    fail3:
    umount("/lib");
    fail4:
#if ${USEPIVOT}
    doHalt();
#endif
    _exit(0);
}
EOF

gcc -Wall -O2 -s -static -nostartfiles -pipe tmp-c-$$.c -o tmp-c-$$
rm -f tmp-c-$$.[co]

x=`cat tmp-c-$$ | wc -c`
y=`expr ${x} + 1023`
x=`expr ${y} / 1024`
y=`expr ${x} + 12`
if [ ${USEGPGKEY} == 1 ] ; then
    y=`expr ${y} + 40`
fi

dd if=/dev/zero of=tmp-i-$$ bs=1024 count=${y}
losetup /dev/loop${DEVFSSLASH2}${TEMPLOOPINDEX} tmp-i-$$
mkfs -t minix -i 32 /dev/loop${DEVFSSLASH2}${TEMPLOOPINDEX}
losetup -d /dev/loop${DEVFSSLASH2}${TEMPLOOPINDEX}
mkdir tmp-d-$$
mount -t minix tmp-i-$$ tmp-d-$$ -o loop=/dev/loop${DEVFSSLASH2}${TEMPLOOPINDEX}
cd tmp-d-$$

mkdir dev lib
mv ../tmp-c-$$ linuxrc

if [ ${USEGPGKEY} == 1 ] ; then
    ln -s lib bin
    ln -s lib/pubring.gpg pubring.gpg
    ln -s lib/secring.gpg secring.gpg
fi

if [ ${USEPIVOT} == 1 ] ; then
    mkdir new-root
else
    mkdir proc
fi

# <device name prefix> <major dev-id> <minor dev-id start> <device 0 suffix>
function maybeMakeDiskNode
{
    x=$3
    for i in "$4" 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 ; do
        for y in ${BOOTDEV} ${CRYPTROOT} ; do
            if [ ${y} == /dev/$1${i} ] ; then
                mknod dev/$1${i} b $2 ${x}
            fi
        done
        x=`expr ${x} + 1`
    done
    return
}
if [ ${USEDEVFS} == 0 ] ; then
    maybeMakeDiskNode hda  3   0 ""
    maybeMakeDiskNode hdb  3  64 ""
    maybeMakeDiskNode hdc 22   0 ""
    maybeMakeDiskNode hdd 22  64 ""
    maybeMakeDiskNode hde 33   0 ""
    maybeMakeDiskNode hdf 33  64 ""
    maybeMakeDiskNode hdg 34   0 ""
    maybeMakeDiskNode hdh 34  64 ""
    maybeMakeDiskNode sda  8   0 ""
    maybeMakeDiskNode sdb  8  16 ""
    maybeMakeDiskNode sdc  8  32 ""
    maybeMakeDiskNode sdd  8  48 ""
    maybeMakeDiskNode sde  8  64 ""
    maybeMakeDiskNode sdf  8  80 ""
    maybeMakeDiskNode sdg  8  96 ""
    maybeMakeDiskNode sdh  8 112 ""
    maybeMakeDiskNode scd 11   0  0
    mknod dev/console c 5 1
    mknod dev/tty c 5 0
    mknod dev/tty1 c 4 1
    mknod dev/null c 1 3
    mknod dev/zero c 1 5
    mknod dev/ram0 b 1 0
    mknod dev/ram1 b 1 1
    mknod dev/loop${ROOTLOOPINDEX} b 7 ${ROOTLOOPINDEX}
fi

cd ..
df tmp-d-$$
umount tmp-d-$$
rmdir tmp-d-$$
sync ; sync ; sync
gzip -9 tmp-i-$$
mv tmp-i-$$.gz ${DESTINATIONROOT}${DESTINATIONPREFIX}/${INITRDGZNAME}
ls -l ${DESTINATIONROOT}${DESTINATIONPREFIX}/${INITRDGZNAME}

if [ ${INITRDONLY} == 1 ] ; then
    echo Done.
    sync
    exit 0
fi

z="/sbin/losetup"
if [ ${USEMODULE} == 1 ] ; then
    z="${z} /sbin/insmod"
fi
if [ ${LOADNATIONALKEYB} == 1 ] ; then
    z="${z} "`which loadkeys`
fi
if [ ${USEGPGKEY} == 1 ] ; then
    z="${z} "`which gpg`
fi
for x in ${z} ; do
    echo Copying ${SOURCEROOT}${x} to ${DESTINATIONROOT}${DESTINATIONPREFIX}
    cp -p ${SOURCEROOT}${x} ${DESTINATIONROOT}${DESTINATIONPREFIX}
    y=`ldd ${SOURCEROOT}${x} | perl -ne 'if(/ => ([^ ]*) /){print "$1\n"}'`
    for a in ${y} ; do
        echo Copying ${SOURCEROOT}${a} to ${DESTINATIONROOT}${DESTINATIONPREFIX}
        cp -p ${SOURCEROOT}${a} ${DESTINATIONROOT}${DESTINATIONPREFIX}
    done
done

if [ ! -d ${DESTINATIONROOT}/initrd ] ; then
    mkdir ${DESTINATIONROOT}/initrd
fi

echo Done.
sync
exit 0
